import shutil
import html.parser

from github import Github

from markdown_parser import MarkdownParser
import mistune
from whoosh.fields import *
import whoosh.index as index
import os
import os.path
import codecs
from whoosh.qparser import MultifieldParser, QueryParser
from whoosh.analysis import StemmingAnalyzer


"""
issues-search.py Flow:


very high level description:
    - zeroth step: create a search index
    - first step: load a search index
    - second step: call the search() method
    - third step: update the search index




program will:
   - create a Search object
   - call add_all_issues
   - calls open_index if no index exists
   - grab a list of repos
   - walk every issue
        - each issue body is (by itself) a document
        - each comment body is also a document
   - call add_document on each issue and each comment
   - add_document parses Markdown/adds docs to index following schema

program will occasionally:
   - ask to update the index
   - this opens the index, and adds all files in the path to a list
   - it then finds files that were deleted and files that were changed
   - it passes these to the indexer, then it's done

program calls the search function:
    - takes query from user
    - calls searcher.search(query, ...)
    - parses results from searcher into user-friendly search results
"""


class SearchResult:
    score = 1.0
    path = None
    content = ""
    content_highlight = ""
    headlines = None
    tags = ""


class DontEscapeHtmlInCodeRenderer(mistune.Renderer):
    def __init__(self, **kwargs):
        super(DontEscapeHtmlInCodeRenderer, self).__init__(**kwargs)

    def block_code(self, code, lang):
        if not lang:
            return '<pre><code>%s\n</code></pre>\n' % code
        return '<pre><code class="lang-%s">%s\n</code></pre>\n' % (lang, code)

    def codespan(self, text):
        return '<code>%s</code>' % text.rstrip()


class Search:
    ix = None
    index_folder = None
    markdown = mistune.Markdown(renderer=DontEscapeHtmlInCodeRenderer(), escape=False)
    html_parser = html.parser.HTMLParser()
    schema = None

    def __init__(self, index_folder):
        self.open_index(index_folder)

    def open_index(self, index_folder, create_new=False):
        self.index_folder = index_folder
        if create_new:
            if os.path.exists(index_folder):
                shutil.rmtree(index_folder)
                print("deleted index folder: " + index_folder)

        if not os.path.exists(index_folder):
            os.mkdir(index_folder)

        exists = index.exists_in(index_folder)
        stemming_analyzer = StemmingAnalyzer()

        schema = Schema(
                url=ID(stored=True, unique=True),
                is_comment=BOOLEAN(stored=True),
                timestamp=STORED,
                repo_name=TEXT(stored=True),
                repo_url=ID(stored=True),
                issue_title=TEXT(stored=True, field_boost=100.0),
                issue_url=ID(stored=True),
                user=TEXT(stored=True),
                content=TEXT(stored=True, analyzer=stemming_analyzer)
        )

        if not exists:
            self.ix = index.create_in(index_folder, schema)
        else:
            self.ix = index.open_dir(index_folder)


    def add_issue(self, writer, issue, repo, config):
        """
        Add Github issue to search index.

        Replaces add_document

        This must:
        - deal with original issue content
        - iterate over each comment
        - deal with comment content

        Schema:
        - url
        - is_comment
        - timestamp
        - repo_name
        - repo_url
        - issue_title
        - issue_url
        - user
        - content
        """

        # should store urls of all issues and comments

        repo_name = repo.name
        repo_url = repo.html_url

        count = 0


        # Handle the issue content
        print("Indexing issue %s"%(issue.html_url))
        writer.add_document(
                url = issue.html_url,
                is_comment = False,
                timestamp = issue.created_at,
                repo_name = repo_name,
                repo_url = repo_url,
                issue_title = issue.title,
                issue_url = issue.html_url,
                user = issue.user.login,
                content = issue.body.rstrip()
        )
        count += 1


        # Handle the comments content
        if(issue.comments>0):
            comments = issue.get_comments()
            for comment in comments:
                print(" > Indexing comment %s"%(comment.html_url))
                writer.add_document(
                        url = comment.html_url,
                        is_comment = True,
                        timestamp = comment.created_at,
                        repo_name = repo_name,
                        repo_url = repo_url,
                        issue_title = issue.title,
                        issue_url = issue.html_url,
                        user = comment.user.login,
                        content = comment.body.strip()
                )

        count += 1
        return count


    '''
    def add_all_issues(self, gh_access_token, list_of_repos, which_org, config, create_new_index=False):
        """
        Add all issues in a given github repo to the search index.

        Replaces add_all_files

        Takes as inputs:
        - github access token
        - list of github repos
        - github org/user owning these repos
        - location of the whoosh config file for configuring the search engine
        """

        if create_new_index:
            self.open_index(self.index_folder, create_new=True)

        writer = self.ix.writer()

        # ------------
        # Iterate over each repo
        # Iterate over each thread (github issue)
        # Iterate over each comment 

        g = Github(gh_access_token)
        org = g.get_organization(which_org)

        c = 0 

        # Iterate over each repo
        for this_repo in list_of_repos:

            repo = org.get_repo(this_repo)
            reponame = repo.name

            # Iterate over each thread
            issues = repo.get_issues()
            for issue in issues:

                # Deal with original issue content
                # AND iterate over each comment 
                c += self.add_issue(writer, issue, repo, config)

                # should store urls of all issues and comments

        writer.commit()
        print("Done, added %d documents to the index" % c)
    '''



    def update_index_incremental(self, gh_access_token, list_of_repos, which_org, config, create_new_index=False):
        """
        Update the index of issues of a given github repo.

        Takes as inputs:
        - github access token
        - list of github repos
        - github org/user owning these repos
        - location of the whoosh config file for configuring the search engine
        """

        if create_new_index:
            self.open_index(self.index_folder, create_new=True)

        # Using URL as the unique identifier
        # Start by getting a list of URLs that are indexed
        # Then walk all URLs


        g = Github(gh_access_token)
        org = g.get_organization(which_org)

        # Set of all URLs as existing on github
        to_index = set()


        writer = self.ix.writer()




        # fix this. the delete all in index
        # is not occurring in right place.


        # Iterate over each repo
        for this_repo in list_of_repos:

            repo = org.get_repo(this_repo)
            reponame = repo.name

            count = 0

            # Iterate over each thread
            issues = repo.get_issues()
            for issue in issues:

                # This approach is more work than is needed
                # but PoC||GTFO

                # For each issue/comment URL,
                # remove the corresponding item
                # and re-add it to the index

                to_index.add(issue.html_url)
                writer.delete_by_term('url', issue.html_url)
                comments = issue.get_comments()

                for comment in comments:
                    to_index.add(comment.html_url)
                    writer.delete_by_term('url', comment.html_url)

                # Now re-add this issue to the index
                count += self.add_issue(writer, issue, repo, config)


        writer.commit()
        print("Done, updated %d documents in the index" % count)


    def create_search_result(self, results):
        # Allow larger fragments
        results.fragmenter.maxchars = 300

        # Show more context before and after
        results.fragmenter.surround = 50

        search_results = []
        for r in results:

            # Note: this is where we package things up 
            # for the Jinja template "search.html".
            # For example, the Jinja template
            # contains a {% for e in entries %}
            # and then an {{e.score}}

            sr = SearchResult()
            sr.score = r.score
            sr.url = r['url']
            sr.title = r['issue_title']

            sr.repo_name = r['repo_name']
            sr.repo_url = r['repo_url']

            sr.issue_title = r['issue_title']
            sr.issue_url = r['issue_url']

            sr.is_comment = r['is_comment']

            sr.content = r['content']
            highlights = r.highlights('content')
            if not highlights:
                # just use the first 1,000 words of the document
                highlights = self.cap(r['content'], 1000)

            highlights = self.html_parser.unescape(highlights)
            html = self.markdown(highlights)
            sr.content_highlight = html
            search_results.append(sr)

        return search_results

    def cap(self, s, l):
        return s if len(s) <= l else s[0:l - 3] + '...'

    def search(self, query_list, fields=None):
        with self.ix.searcher() as searcher:
            query_string = " ".join(query_list)
            query = None
            if "\"" in query_string or ":" in query_string:
                query = QueryParser("content", self.schema).parse(query_string)
            elif len(fields) == 1 and fields[0] == "filename":
                pass
            elif len(fields) == 2:
                pass
            else:
                fields = ["headlines", "content", "filename", "doubleemphasiswords", "emphasiswords"]
            if not query:
                query = MultifieldParser(fields, schema=self.ix.schema).parse(query_string)
            parsed_query = "%s" % query
            print("query: %s" % parsed_query)
            results = searcher.search(query, terms=False, scored=True, groupedby="url")
            search_result = self.create_search_result(results)

        return parsed_query, search_result

    def get_document_total_count(self):
        return self.ix.searcher().doc_count_all()

if __name__ == "__main__":
    search = Search("search_index")

    repo_list = ['2018-may-workshop',  
                 '2018-june-workshop', 
                 '2018-july-workshop'] 
    repo_org = 'dcppc'

    gh_access_token = os.environ['GITHUB_TOKEN']

    search.add_all_issues(gh_access_token,
                          repo_list,
                          repo_org,
                          "/Users/charles/codes/issues-search/config.py")

